# Built-in types and functions exposed to users

int64 = None
int32 = None
float64 = None
float32 = None

def Array(ty, ndim):
    return

def List(ty):
    return

def get_row(a, i):
    return a[i]
